""" The application module provides an interface for WSGI servers and other important functions specified in the App class."""


from wsgiref.simple_server import make_server
from .parser import  Context
from . import router
from .http_response import response, static_response, load



class App(object):
    """The App class encapsulates the neccessary functions needed to make the framework compatible with WSGI servers. This is basically the backbone of Webkage.
    The self.wsgi is the interface by which a Webkage app can be served by WSGI compatible servers. In development, the self.serve function can be used in place of a standard WSGI server."""
    def __init__(self):
        self.routes_func = []
        self.routes_no_func = []
        self.static_path_prefix = '--'
        self.static_dir = ''


    def match_route(self, context):
        """This method aims to differentiate between requests for static files vs others. You won't and shouldn't invoke this method."""

        global resp


        path = context.request["path"]
        #Check if static
        if path.startswith(self.static_path_prefix):
            resp = self.serve_static(context, path)
        else:
            if path[-1] != "/":
                path += "/"
            resp = self.serve_response(context, path)

        return resp
       


    def serve_response(self, context, path):
        """This method return a HTTP response back to the client.
        It checks if the requested path matches with any path registered using self.add_path. If none, a 404 page is served using self.not_found. """
        global response_

        matched_route = router.match_route(self.routes_no_func, path)
        if matched_route == None:
            response_ = not_found(context)
        else:
            id_, slug = router.get_params(matched_route, path)
            context.params["slug"] = slug
            context.params["id"] = id_
            for route in self.routes_func:
                if route[0] == matched_route:
                    response_ = route[1](context) #Execute view function
        return response_ 




    def wsgi(self, environ, start_response):
        """This is the WSGI main entry point. All WSGI servers should reference this point.

        Here's an example with Gunicorn:

        #main.py

        from webkage.application import App
        from webkage.http_response import response

        ...

        app = App
        wsgi = app.wsgi

        $gunicorn main:app.wsgi
        """

        context_ = Context(environ)
        context_.response["start_response"] = start_response
        response = self.match_route(context_)
        yield response


    def set_static(self, static_path_prefix, static_dir):
        """The first argument specifies the prefix of all static files while the second argument specifies which directory to serve them from."""
        self.static_path_prefix = static_path_prefix
        if static_dir[-1] != '/':
            self.static_dir = static_dir + '/'


    def serve_static(self, context, path):
        """Unlike self.serve_response, this method serves only static files to the client."""
        file_path = self.static_dir + path.replace(self.static_path_prefix, '').rstrip("/")
        status = "200 OK"
        try:
            with open(file_path, "rb") as file_:
                static_file = file_.read()
            if file_path.endswith(".css"):
                context.request["content-type"] = "text/css"
        except Exception as e:
            static_file = b''
            status = "404 NOT FOUND"
        return static_response(context, status, static_file)


    def add_path(self, uri, method):
        """Use this method to register all paths that should be discoverable by Webkage. Simple paths or complicated ones with parameters can be specified.

        #main.py

        ...

        app = App


        def home(ctx):
            ...

        app.add_path("/home", home)

        ...

        """

        if uri[-1] != "/":
            uri += "/"
        uri = router.path_to_regex(uri)
        path = (uri, method)
        self.routes_func.append(path)
        self.routes_no_func.append(uri)


    def serve(self, host="127.0.0.1", port=8000):
        """Serve App locally. Avoid using it in production, should be disabled before deployment."""

        httpd = make_server(host, port, self.wsgi)
        print("Listening on %s:%d" %(host, port))
        httpd.serve_forever()




def not_found(context):
    """ 404 response view """
    try:
        data = load("404.html")
    except:
        data = b"""
        <html>
            <head>
                    <title>404 Not Found</title>
            </head>
            <body>
            <h2>404 Not Found</h2>
            </body>
        </html>
        """
    return response(context, "404 NOT FOUND", data)





"""This module handles and parses http responses. """

from .parser import set_cookie_header
import json
import jinja2



def response(context, status, data):
    """This method should be returned at the end of your views. All three arguments are compulsory and must be specified.


    #main.py
    ...

    def home(ctx):
        ...
        return response(ctx, ""200 OK", myData)

    ...

    """
    content_type = "text/html"
    set_cookie = set_cookie_header(context.session)
    header = [("Content-Type", content_type),] + set_cookie
    start_response = context.response["start_response"]
    start_response(status+" ", header)
    return data


def json_response(context, status, data):
    """Pure JSON objects can be returned as a response by invoking this method. The second argument should be a Pythonic object that can be converted to JSON by the json library."""
    content_type = "application/json"
    set_cookie = set_cookie_header(context.session)
    header = [("Content-Type", content_type),] + set_cookie
    start_response = context.response["start_response"]
    start_response(status+" ", header)
    return jsonify(data)


def static_response(context, status, data):
    """A dummy method that serves static files to the client. This function is invoked by the App module to serve static files."""
    content_type = context.request["content-type"]
    header = [("Content-Type", content_type),]
    start_response = context.response["start_response"]
    start_response(status, header)
    return data


def load(path, data=""):
    """"This method searches for template(HTML) files under ./templates directory relative to where the python file that contains the WSGI entry point lies. Data are enconded and rendered by Jinja2."""
    loader = jinja2.FileSystemLoader(searchpath="./templates")
    env = jinja2.Environment(loader=loader)
    template = env.get_template(path)
    data_ = template.render(data).encode("utf-8")
    return data_


def jsonify(obj):
    """Return Python objects as JSON (in bytes)."""
    jsonified = "{}".format(json.dumps(obj)).encode()
    return jsonified


def redirect(context, status, uri):
    """A HTTP redirect can be invoked by calling this method.


    #main.py
    ...

    def home(ctx):
        ...
        return redirect(ctx, "301 Redirect", "/dashboard")
    """
    header = [("Location", uri)]
    start_response = context.response["start_response"]
    start_response(status+" ", header)
    return b""





""" Router module for matching routes/paths."""
import re


def path_to_regex(path):
    path_segments = path.split("/")
    params = dict()
    for index, segment in enumerate(path_segments):
        if segment == ":id":
            path_segments[index] = "[0-9]+"
        elif segment == ":slug":
            path_segments[index] = "[a-zA-Z]+(?:-[a-zA-Z]+)*"
    path = "/".join(path_segments) + "$"
    return path



def match_route(routes, path):
    for route in routes:
        match = re.match(route, path)
        if match is not None:
            return route


def get_params(match, path):
    """URI parameters are parsed here."""
    path_segments = path.split("/")
    match_segments = match.split("/")
    id_ = None
    slug = None
    id_index = None
    slug_index = None
    if "[0-9]+" in match_segments:
        id_index = match_segments.index("[0-9]+")
    if "[a-zA-Z]+(?:-[a-zA-Z]+)*" in match_segments:
        slug_index = match_segments.index("[a-zA-Z]+(?:-[a-zA-Z]+)*")
    if id_index != None:
        id_ = path_segments[id_index]
    if slug_index != None:
        slug = path_segments[slug_index]

    return (id_, slug)






        
        








""" Parser module to parse environ variables."""

from http.cookies import SimpleCookie
import io
import cgi
import re
import json


def clean(item):
    if item != "":
        return True

def set_cookie_header(cookie):
    """Parse cookie object to HTTP Set-Cookie header"""

    header = list()
    cookie_options = cookie.output().split("Set-Cookie:")
    cookie_options = list(filter(clean, cookie_options))
    for index, option in enumerate(cookie_options):
        cookie_options[index] = option.strip("\r\n")
    cookie_options = ",".join(cookie_options) + '; ' + ';'.join(cookie.options)
    

    header = [("Set-Cookie", cookie_options),]
    return header




class Cookie(SimpleCookie):
    """Cookie class inheriting http.SimpleCookie class"""

    def __init__(self):
        super().__init__()
        self.options = ["HttpOnly",]

    def __setitem__(self, key, value):
        super().__setitem__(key, value)


    def httponly(self, choice):
        """Set httponly option in Set-Cookie"""
        if choice == True:
            if not "HttpOnly" in self.options:
                self.options.append("HttpOnly")
        elif choice == False:
            if "HttpOnly" in self.options:
                self.options.remove("HttpOnly")

    def secure(self, choice):
        if choice == True:
            if not "Secure" in self.options:
                self.options.append("Secure")
        elif choice == False:
            if "Secure" in self.options:
                self.options.remove("Secure")
    
    def flush(self):
        """Expires a given cookie value"""

        self.options.append("expires=Mon 23 Jun 1967 04:34:23 GMT")

    def will_expire(self, date):
        """Sets future expiry date for cookie value"""

        self.options.append("expires={}".format(date))



class Context():
    """Make needed environ variables accessible in a single class"""

    def __init__(self, environ):

        #Get incoming cookies
        cookie = environ.get("HTTP_COOKIE", "")
        self.session = Cookie()
        self.session.load(cookie)


        #HTTP Request
        self.request = dict()
        self.request["content-length"] = environ.get("CONTENT_LENGTH", 0)
        self.request["content-type"] = environ.get("CONTENT_TYPE", "text/html")
        self.request["user-agent"] = environ.get("HTTP_USER-AGENT", "")
        self.request["ip-address"] = environ.get("REMOTE_ADDR", "")
        self.request["path"] = environ.get("PATH_INFO", "/")
        self.query_str = environ.get("QUERY_STRING")
        self.request["method"] = environ.get("REQUEST_METHOD")
        self.request["protocol"] = environ.get("SERVER_PROTOCOL", "")
        self.request["scheme"] = environ.get("wsgi.url_scheme", "http")
        self.request["data"] = environ.get("wsgi.input", "")


        #HTTP Response dict
        self.response = dict()

        #HTTP Parameters dict()
        self.params = dict()

        if self.request["content-length"] == '':
            self.request["content-length"]=0

        #HTTP Form
        #Parse form content
        self.form = dict()
        self.formFile = dict()
        self.json = ''
        fp = self.request["data"].read(int(self.request["content-length"]))
        if self.request["content-type"] == "application/x-www-form-urlencoded" or self.request["content-type"] == "multipart/form-data" :
            formValues = cgi.FieldStorage(fp=io.BytesIO(fp), environ=environ, keep_blank_values=True)
            keys = formValues.keys()
            for key in keys:
                value = formValues[key]
                if value.filename:
                    self.formFile[key] = value
                else:
                    self.form[key] = formValues.getvalue(key)
        elif self.request["content-type"] == "application/json":
            #Process Json request
            self.json = json.loads(fp.decode())


        
        
        #Query matching and parsing
        self.query = dict()

        query_regex = "[a-zA-Z0-9%+]+=[a-zA-Z0-9%+]+"
        queries = re.findall(query_regex, self.query_str)
        for index, query in enumerate(queries):
            queries[index] = query.replace("%20", " ").replace("+", " ")
        
        #Populate Query dict
        for query in queries:
            q = query.split("=")
            self.query[q[0]] = q[1]



